<?php

namespace Drupal\facets_date_range_picker\Plugin\facets\query_type;

use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\facets\QueryType\QueryTypeRangeBase;
use Drupal\facets\Result\Result;
use Drupal\search_api\Query\ConditionGroupInterface;
use Drupal\search_api\Query\QueryInterface;

/**
 * Support for date range facets within the Search API scope.
 *
 * This query type supports dates for all possible backends. This specific
 * implementation of the query type supports a generic solution of adding facets
 * for dates.
 *
 * If you want to have a specific solution for your backend / module to
 * implement dates, you can alter the ::getQueryTypesForDataType method on the
 * backendPlugin to return a different class.
 *
 * @FacetsQueryType(
 *   id = "search_api_date_range_picker",
 *   label = @Translation("Date Range Picker"),
 * )
 */
class SearchApiDateRangePicker extends QueryTypeRangeBase {

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $facet = $configuration['facet'];
    $processors = $facet->getProcessors();
    $dateProcessorConfig = $processors['date_range_picker']->getConfiguration();

    $configuration = $this->getConfiguration();
    $configuration['start_date'] = $dateProcessorConfig['start_date'];
    $configuration['end_date'] = $dateProcessorConfig['end_date'];
    $configuration['options'] = $dateProcessorConfig['options'];
    $configuration['overlap'] = $dateProcessorConfig['overlap'];
    $this->setConfiguration($configuration);
  }

  /**
   * {@inheritdoc}
   */
  public function execute() {
    $query = $this->query;

    // Alter the query here.
    if (!empty($query)) {
      $options = &$query->getOptions();

      $operator = $this->facet->getQueryOperator();
      $field_identifier = $this->facet->getFieldIdentifier();
      $exclude = $this->facet->getExclude();
      $options['search_api_facets'][$field_identifier] = $this->getFacetOptions();

      // Add the filter to the query if there are active values.
      $active_items = $this->facet->getActiveItems();
      $filter = $query->createConditionGroup($operator, ['facet:' . $field_identifier]);
      if (count($active_items)) {
        foreach ($active_items as $value) {
          if ($this->getOverlap()) {
            $this->addConditionGroupsByActiveItemWithOverlap($query, $filter, $exclude, $value, $field_identifier);
          }
          else {
            $this->addConditionGroupsByActiveItem($query, $filter, $exclude, $value, $field_identifier);
          }
        }
        $query->addConditionGroup($filter);
      }
      else {
        $range = $this->calculateDefault();

        $conjunction = $exclude ? 'OR' : 'AND';
        $item_filter = $query->createConditionGroup($conjunction, ['facet:' . $field_identifier]);
        $item_filter->addCondition($this->facet->getFieldIdentifier(), $range['start'], $exclude ? '<' : '>=');
        $filter->addConditionGroup($item_filter);

        $query->addConditionGroup($filter);
      }
    }
  }

  /**
   * Create the condition group for the active item.
   *
   * @todo Allow to exclude items.
   *
   * @param \Drupal\search_api\Query\QueryInterface $query
   *   The query.
   * @param \Drupal\search_api\Query\ConditionGroupInterface $filter
   *   The filter.
   * @param bool $exclude
   *   Whether or not to exclude the items.
   * @param string $active_item
   *   The active item.
   * @param string $field_identifier
   *   The field identifier.
   */
  protected function addConditionGroupsByActiveItem(QueryInterface &$query, ConditionGroupInterface &$filter, bool $exclude, string $active_item, string $field_identifier) {
    $end_date = $this->getEndDate();
    $range = $this->calculateRange($active_item);

    $item_filter = $query->createConditionGroup('AND', ['facet:' . $field_identifier]);
    $item_filter->addCondition($end_date, $range['start'], $exclude ? '<' : '>=');
    $item_filter->addCondition($end_date, $range['stop'], $exclude ? '>' : '<=');

    $filter->addConditionGroup($item_filter);
  }

  /**
   * Create the overlapping condition group for the active item.
   *
   * @todo Allow to exclude items.
   *
   * @param \Drupal\search_api\Query\QueryInterface $query
   *   The query.
   * @param \Drupal\search_api\Query\ConditionGroupInterface $filter
   *   The filter.
   * @param bool $exclude
   *   Whether or not to exclude the items.
   * @param string $active_item
   *   The active item.
   * @param string $field_identifier
   *   The field identifier.
   */
  protected function addConditionGroupsByActiveItemWithOverlap(QueryInterface &$query, ConditionGroupInterface &$filter, bool $exclude, string $active_item, string $field_identifier) {
    $start_date = $this->getStartDate();
    $end_date = $this->getEndDate();
    $range = $this->calculateRange($active_item);

    $item_filter = $query->createConditionGroup('OR', ['facet:' . $field_identifier]);

    // Events starting and stopping between range.
    $item_sub_filter_1 = $query->createConditionGroup('AND', ['facet:' . $field_identifier . '_1']);
    $item_sub_filter_1->addCondition($start_date, $range['start'], '>=');
    $item_sub_filter_1->addCondition($start_date, $range['stop'], '<=');
    $item_filter->addConditionGroup($item_sub_filter_1);

    // Events started before and stop between range.
    $item_sub_filter_2 = $query->createConditionGroup('AND', ['facet:' . $field_identifier . '_2']);
    $item_sub_filter_2->addCondition($start_date, $range['start'], '<=');
    $item_sub_filter_2->addCondition($end_date, $range['start'], '>=');
    $item_sub_filter_2->addCondition($end_date, $range['stop'], '<=');
    $item_filter->addConditionGroup($item_sub_filter_2);

    // Events started before and stop after range.
    $item_sub_filter_3 = $query->createConditionGroup('AND', ['facet:' . $field_identifier . '_3']);
    $item_sub_filter_3->addCondition($start_date, $range['start'], '<=');
    $item_sub_filter_3->addCondition($end_date, $range['stop'], '>=');
    $item_filter->addConditionGroup($item_sub_filter_3);

    $filter->addConditionGroup($item_filter);
  }

  /**
   * Calculate the default for a given facet filter value.
   *
   * Used when adding active items in self::execute() to $this->query to include
   * the range conditions for the value.
   *
   * @return array
   *   Keyed with 'start' values.
   */
  protected function calculateDefault(): array {
    $dateTime = new DrupalDateTime();
    $date = new DrupalDateTime('now');

    $startDate = $dateTime::createFromFormat('Y-m-d\TH:i:s', $date->format('Y-m-d\TH:i:s'));

    return [
      'start' => $startDate->format('U'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function calculateRange($value) {
    $dateTime = new DrupalDateTime('now', 'UTC');
    $now = $dateTime;
    $now->setTime(0, 0);

    switch ($value) {
      case 'today':
        $start_date = $end_date = new DrupalDateTime('now', 'UTC');
        break;

      case 'tomorrow':
        $start_date = $end_date = new DrupalDateTime('tomorrow', 'UTC');
        break;

      case 'this_weekend':

        // Today is saturday.
        if ($now->format('N') === '6') {
          $start_date = $now;
          $end_date = $now->modify('+1 day');
        }
        // Today is sunday.
        elseif ($now->format('N') === '7') {
          $start_date = $now;
          $end_date = $now->modify('-1 day');
        }
        // Today is just a weekday.
        else {
          $start_date = new DrupalDateTime('next saturday', 'UTC');
          $end_date = new DrupalDateTime('next sunday', 'UTC');
        }
        break;

      case 'next_seven_days':
        $start_date = new DrupalDateTime('now', 'UTC');
        $end_date = new DrupalDateTime('+7 days', 'UTC');
        break;

      case 'next_fourteen_days':
        $start_date = new DrupalDateTime('now', 'UTC');
        $end_date = new DrupalDateTime('+14 days', 'UTC');
        break;

      case 'next_thirty_days':
        $start_date = new DrupalDateTime('now', 'UTC');
        $end_date = new DrupalDateTime('+30 days', 'UTC');
        break;
    }

    $startDate = $dateTime::createFromFormat('Y-m-d\TH:i:s', $start_date->format('Y-m-d') . 'T00:00:00');
    $stopDate = $dateTime::createFromFormat('Y-m-d\TH:i:s', $end_date->format('Y-m-d') . 'T23:59:59');

    return [
      'start' => $startDate->format('U'),
      'stop' => $stopDate->format('U'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    // If there were no results or no query object, we can't do anything.
    if (empty($this->results)) {
      return $this->facet;
    }

    $facet_results = [];
    foreach ($this->getOptions() as $key => $option) {
      if ($option) {
        $result_filter = $this->calculateResultFilter($option);
        $facet_results[$key] = new Result($this->facet, $result_filter['raw'], $result_filter['display'], 0);
      }
    }

    $this->facet->setResults($facet_results);
    return $this->facet;
  }

  /**
   * Calculates the result of the filter.
   *
   * @param string $value
   *   The key.
   *
   * @return array
   *   An array with the display and the raw value.
   */
  public function calculateResultFilter($value) {

    switch ($value) {
      case 'today':
        $display = $this->t('Today');
        break;

      case 'tomorrow':
        $display = $this->t('Tomorrow');
        break;

      case 'this_weekend':
        $display = $this->t('This weekend');
        break;

      case 'next_seven_days':
        $display = $this->t('Next 7 days');
        break;

      case 'next_fourteen_days':
        $display = $this->t('Next 14 days');
        break;

      case 'next_thirty_days':
        $display = $this->t('Next 30 days');
        break;
    }

    return [
      'display' => $display,
      'raw' => $value,
    ];
  }

  /**
   * Retrieve configuration: Start date field.
   *
   * @return string
   *   The start date for this config.
   */
  protected function getStartDate(): string {
    return $this->getConfiguration()['start_date'];
  }

  /**
   * Retrieve configuration: End date field.
   *
   * @return string
   *   The end date for this config.
   */
  protected function getEndDate(): string {
    return $this->getConfiguration()['end_date'];
  }

  /**
   * Retrieve configuration: Options to display.
   *
   * @return string[]
   *   The options for this config.
   */
  protected function getOptions(): array {
    return $this->getConfiguration()['options'];
  }

  /**
   * Retrieve configuration: Overlap setting.
   *
   * @return bool
   *   The overlap setting for this config.
   */
  protected function getOverlap(): bool {
    return $this->getConfiguration()['overlap'];
  }

}
